#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place
# - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2001-2005 Free Software Foundation
#
# FILE:
# xmlrpc/ServerAdapter.py
#
# DESCRIPTION:
# Set of classes that implement the XML-RPC server driver for GNUe Comm.
#
# NOTES:
# Requires xmlrpclib from http://www.pythonware.com/products/xmlrpc/
# or Python 2.2+
#
# Server Parameters:
#
#    port        The port that the service is located on
#

from gnue.common.rpc import server
from gnue.common.apps import errors
from gnue.common.rpc.drivers._helpers import ObjectLibrarian, DirectoryServer

from BaseHTTPServer import BaseHTTPRequestHandler;
from SimpleHTTPServer import SimpleHTTPRequestHandler;
import SocketServer

import string, sys, os, posixpath, urllib, socket, types

try:
  import xmlrpclib
except ImportError:
  tmsg = _("\nUnable to load xmlrpclib.  To use the XML-RPC interface, \n"
           "please install xmlrpc from:\n"
           "    http://www.pythonware.com/products/xmlrpc/")
  raise server.AdapterInitializationError, tmsg

import typeconv

# Mapping from GRPC's datatype to XML-RPC datatypes
_datatypeMap = {
  'integer': 'int',
  'string': 'string',
  'boolean': 'boolean',
  'date': 'dateTime.iso8601',
  'number': 'double',
  'base64': 'base64',
  'binary': 'base64'
}

class MyThreadingHTTPServer(SocketServer.ThreadingTCPServer):
  def setParent(self,parent):
    self._parent=parent
    
  def server_bind(self):
    self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    self.socket.bind(self.server_address)

class MyForkingHTTPServer(SocketServer.ForkingTCPServer):
  def setParent(self,parent):
    self._parent=parent
    
  def server_bind(self):
    self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    self.socket.bind(self.server_address)


# TO BE CHECKED:
# Python 2.1: SocketServer.ThreadingMixIn
#SocketServer.ThreadingMixIn does not work properly
#since it tries to close the socket of a request two
#times.

#Workaround for using SocketServer.ThreadingMixIn under
#Python 2.1:

#class MyThreadingHTTPServer(
#SocketServer.ThreadingMixIn,
#MyHTTPServer
#):
#def close_request(self, request):
#pass

# an medusa like server 
#  should also be implemented: import asyncore, overwrite asyncore.dispatcher


##############################################################################
#
# ServerAdapter
#
class ServerAdapter(DirectoryServer.DirectoryServer):

  def __init__(self, rpcdef, bindings, params):
    try:
      self._port = params['port']
    except KeyError:
      tmsg = _('Required parameter "port" not supplied') 
      raise server.AdapterConfigurationError, tmsg 

    if params.has_key('bindto'):
      self._bindto = params['bindto']
    else:
      self._bindto = '' # bind to all interfaces

    if params.has_key('allowed_hosts'):
      # TODO: Remove spaces, etc.
      self._allowed_hosts = string.split(params['allowed_hosts'],',')
    else:
      self._allowed_hosts = [''] # allow access from all hosts

    if params.has_key('loglevel'):
      self._loglevel = params['loglevel']
    else:
      self._loglevel = 0

    if params.has_key('httpbind'):
      self._httpbind = params['httpbind']
      
    self._httpserver=MyThreadingHTTPServer
    
    if params.has_key('servertype'):
      if params['servertype']=='threading':
        self._httpserver=MyThreadingHTTPServer            
      elif params['servertype']=='forking':
        self._httpserver=MyForkingHTTPServer
#      elif params['servertype']=='asyncore':
#        self._httpserver=MyAsyncoreDispatcher
      else:
        raise server.AdapterConfigurationError, \
              u_("Value %(value)s for parameter 'servertype' is not supported. "
                 "Valid values are: %(valid)s") \
              % {'value': params ['servertype'],
                 'valid': 'threading, forking'}

    DirectoryServer.DirectoryServer.__init__(self, rpcdef, bindings, params)
    
    # To make "Object" definitions with w/non-object RPC,
    # the first parameters to any "Object" methods are
    # a "passed reference"
    # if parent._type == 'RpObject':
    #    signature.append(_datatypeMap['string'])
    
    # To simulate "Object" definitions with w/this
    # non-object RPC, the first parameters to any
    # "Object" methods are a "passed reference"

  #
  # Return an exception
  #
  def raiseException(self, exception, message, event=None):
    xmlrpclib.dumps(xmlrpclib.Fault(34543, '%s: %s' % (exception, message)))



  #                                                                        #
  #                                                                        #
  ##########################################################################
  def serve(self):
    if hasattr(self,'_httpbind'):
      self._requestHandler = EnhancedRequestHandler
    else:
      self._requestHandler = RequestHandler
      
    self._tcpserver = self._httpserver((self._bindto,self._port),
                                       self._requestHandler)
    self._tcpserver.setParent(self)
    try: 
      self._tcpserver.serve_forever()
    except KeyboardInterrupt:
      pass
    
  #
  # Call the requested method
  #

  # ---------------------------------------------------------------------------
  # Call the requested method
  # ---------------------------------------------------------------------------

  def call (self, method, params):
    if self._loglevel > 0:
      print _("Dispatching: "), method, params
    

    ## Check if the Method is part of a service or a pointer to a
    ## single object
    ##
    ## Call to an object:  method="_Management:235423456_.getAttr"
    ##                     params=""
    ## Call to an method: (of a service=one object)
    ##                     method="DonutPlace.Management.Restart"
    if method [0] == '[':
      # call to an object
      # 1. get the object from the objectlibrarian
      # 2. check, if the object is supported by the gfd

      try:
        i = string.index (method, ']', 1)
        objhandle = method [1:i]
        method = method [i+2:]

      except ValueError:
        raise errors.ApplicationError, \
            u_("Wrong format of object handle in method call %s") % method

      # TODO check in service dir, if obj is supported or not
      o = ObjectLibrarian.retrieveObject (objhandle)

      try:
        server_method = getattr (o, method)
        server_attribute = None

      except AttributeError:
        server_method = None

        try:
          server_attribute = getattr (o, method [4:])

        except AttributeError:
          if method != "_close":
            raise errors.ApplicationError, \
                u_("Internal XMLRPC server error: method %s can be "
                   "found in the directory (build out of a .grpc file), "
                   "but the object doesn't contain this method/attribut. "
                   "Please check you .grpc file for wrong return types.") \
                 % method
        
      if method != "_close":
        direntry  = self.getMethodDirEntry (o._type + "." + method)
        signature = direntry ['signature']
      else:
        signature = ('string',)

    else:
      # call to a service method or a helper call (get/set) for
      # a service attribut
      direntry         = self.getMethodDirEntry (method)
      server_method    = direntry ['binding']
      server_attribute = None
        
      # check if it is an real method (binding -> method)
      # or an get/set method for an attribut (binding-> attribut)

      if (type (server_method) != type (self.call)):
        server_attribute = server_method
        server_method    = None
          
      signature = direntry ['signature']

      if (server_method is None) and (server_attribute is None):
        raise errors.SystemError, \
            u_("Server XML-RPC method '%s' is not bound to real method") \
            % method

    try:
      # TODO:  Compare submitted attributs with signature
      pass

    except KeyError:
      raise errors.ApplicationError, \
          u_("Server XML-RPC procedure %(method)s accepts just %(attr)s "
             "as attributes") \
          % {'method': method,
             'attr'  : attr}
    

    # replace object handles in param with the real object
    counter = 0
    while counter < len (params):
      p = params [counter]
      if type (p) == type (""):
        if (len (p) == 42) and (p [0] == "[") and (p [41] == "]"):
          try:
            p = p [1:41]
            obj = ObjectLibrarian.retrieveObject (p)
            newp = params [0:counter-1] + (obj,) + params [counter+1:]
            params = newp

          except:
            pass

      counter += 1

    # check if it is an real method (binding -> method)
    # or an get/set method for an attribut (binding-> attribut)

    __params = typeconv.rpc_to_python (params, server.InvalidParameter)

    if (server_method is not None):
            
      # call method with params
      try:
        result = server_method (*__params)
      except:
        # Conserve traceback
        (group, name, message, detail) = errors.getException (1)
        raise errors.RemoteError (group, name, message, detail)

      result = typeconv.python_to_rpc (result, server.InvalidParameter)

    # check if it's the close object method
    elif (method == "_close"):
      result = self.releaseDynamicObject (o)
      
    else:
      
      ## check wether it's the set or the get method for the attribute
      mparts = string.splitfields (method, '.')
      mparts.reverse ()
      calltype = mparts [0]
      calltype = calltype [:4]
      gDebug (9, 'method %s has calling type %s' % (method, calltype))

      if calltype == 'set_':
        # setAttribut method
        server_attribute = params [0]

      elif calltype == 'get_':
        # getAttribut method
        result = server_attribute
      else:
        raise errors.SystemError, \
            u_("Internal Server XML-RPC error: method type (get/set "
               "attribute) couldn't be detected (method %s)") % method
    

    # replace real object in param with an object handle
    if type (result) == type (self):  ## both should be instances
       ObjectLibrarian.archiveObject (result)

       # get the type of the result
       rtype = signature [0]
       # delete the surrounding brackets < >
       rtype = rtype [1:len(rtype)-1]
       # store typeinfo in new object
       result._type = rtype

       result = ObjectLibrarian.getObjectReference (result)
       self.registerDynamicObject (result, rtype)

    # check for empty results (not allowed for XMLRPC)
    if (result is None) or (result == [None]):
      gDebug (9, 'Transform result None into 1')
      result = 1

    return result
  
  
  def registerDynamicObject(self,objhandle,type):
    #
    #  add new methods to the xmlrpc server
    #
    tl=len(type)
    for i in self.directory.keys():
      if i[0:tl]==type:
        method="["+objhandle+"]"+i[tl:]
        gDebug (9, 'Method %s registered to py-xmlrpc internal directory.' \
                   % method)
 #       self.server.addMethods({method:None})
     
    # add a method to close this object
  #  self.server.addMethods({"["+objhandle+"]._close":None})

  def deregisterDynamicObject(self,object):
    #
    #  remove the new methods from the xmlrpc server
    #
    objhandle=ObjectLibrarian.getObjectReference(object)
    type=object._type
    tl=len(type)
    for i in self.directory.keys():
      if i[0:tl]==type:
        method="["+objhandle+"]"+i[tl:]
        gDebug (9, 'Method %s is deleted from py-xmlrpc ' \
                         % method +\
                         ' internal directory.')
#        self.server.removeMethods({method:None})

    # remove the method to close this object
#    self.server.removeMethods({"["+objhandle+"]._close":None}) 

  def releaseDynamicObject(self,object):
    # delete bindings for this object in the xmlrpc method directory
    #    self.deregisterDynamicObject(object)
    # release object for Object Librarian
    ObjectLibrarian.deferenceObject(object)
    



##############################################################################
#
# Class to Handle the HTTP Post Request for XMLRPC only
#

class RequestHandler(BaseHTTPRequestHandler):

  def log_message (self, format, *args):
    gDebug (9, "%s - - [%s] %s" % \
        (self.address_string(), self.log_date_time_string(), format%args))

  # override basic handle to test allowed host condition
  #
  def handle(self):
    for i in self.server._parent._allowed_hosts:
      # get ip address of client and compare its first characters with i
      if self.client_address[0][:len(i)]==i:
        
        # continue with normal processing
        return BaseHTTPRequestHandler.handle(self)

    self.requestline="UNKNOWN"
    self.request_version="UNKNOWN"
    self.send_error(403, "Access to this server is forbidden!")
    return
      
    
  #
  # Handle the HTTP POST command
  #
  def do_POST(self):
    try:
      # Read the request
      data = self.rfile.read (int (self.headers ["content-length"]))
      params, method = xmlrpclib.loads (data)
      
      # Execute the function
      response = self.server._parent.call (method, params)

      # Generate response as a tuple
      if response is None:
        response = ()
      elif not isinstance (response, types.TupleType):
        response = (response,)
      response = xmlrpclib.dumps (response, methodresponse = 1)
    except:
      # Send back exceptions as Fault responses
      stack  = string.join (errors.getException (), u'\x91')
      response = xmlrpclib.Fault (1, stack)
      response = xmlrpclib.dumps (response, methodresponse = 1)

    self.send_response (200)
    self.send_header ("Content-type", "text/xml")
    self.send_header ("Content-length", str (len (response)))
    self.end_headers ()
    self.wfile.write (response)

    # shut down the connection
    self.wfile.flush ()
    self.connection.shutdown (1)


##############################################################################
#
# Enhanced HTTP Request Handler serving XMLRPC Post requests and allows to
# access appserver data directory
#

class EnhancedRequestHandler(RequestHandler,SimpleHTTPRequestHandler):

  def send_head(self):
    # check for handler providing file content in string form
    try:
      # TODO: rewrite, add a hook instead of just filtering single pages
      content=self.server._parent._httpbind[self.path]()
      self.send_response(200)
      self.send_header("Content-type", "text/html")
      self.send_header("Content-Length", str(len(content)))
      self.end_headers()
      if self.command=="GET":
        self.wfile.write(content)
        self.wfile.flush()
      return None
    except:
      return SimpleHTTPRequestHandler.send_head(self)

  
  def translate_path(self, path):
    path = posixpath.normpath(urllib.unquote(path))
    words = path.split('/')
    words = filter(None, words)    
    httpbind = self.server._parent._httpbind
    
    try:
      path=httpbind["/"]
    except:
      path=httpbind

    ## security check      
    if (type(path)!=type("") or (len(path)<4)):
      return "/dev/null"
    
    for word in words:
      drive, word = os.path.splitdrive(word)
      head, word = os.path.split(word)
      if word in (os.curdir, os.pardir): continue
      path = os.path.join(path, word)
      
    return path
    

###########################
# The following part is not used at the moment
##############################################################################
#
# Used internally by the XML-RPC server to proxy
# between Object-by-Ref and actual objects.
#
class _ObjectByReferenceHandler:

  def __init__(self, oclass):
    self._class = oclass

  # If this makes absolutely no sense, move on and trust the code :)
  def requestHandler(self, method, *params, **args):

    if len(params):

      # If parameters were passed, the first one is always the handle
      refid = params[0]

      # The object's real method doesn't expect to receive the object handle
      params = params[1:]

    else:

      try:

        # If only named parameters are present, then
        # obj_handle should be the object handle...
        refid = args['obj_handle']

        # The object's real method doesn't expect to receive the object handle
        del args['obj_handle']

      except KeyError:
        tmsg = _('Object handle not returned')
        raise StandardError, tmsg   # TODO

    try:
      return ObjectLibrarian. \
        retrieveObject(refid).__dict__[method](*params, **args)
    except KeyError:

      # Attempt to use an invalid objecthandle
      tmsg = _('Invalid object handle')
      raise StandardError, tmsg # TODO



